"""
This module defines scales and some functions which can operate on scales.
"""

__author__ = "https://codeberg.org/decoherence"

from collections import deque
from copy import deepcopy
from itertools import accumulate

from .chords import IDScaleDegree, IDChord, CHORD_TYPES
from .notes import note_name, note_to_note_value, normalize_note_value, normalize_note_values, note_values_to_note_names
from .tone_intervals import IDToneInterval, transpose_tone_intervals


class IDScaleIntervals:

    """Defines the intervals for the different scale genus."""
    DIATONIC = [
        IDToneInterval.R,
        IDToneInterval.M2,
        IDToneInterval.M3,
        IDToneInterval.P4,
        IDToneInterval.P5,
        IDToneInterval.M6,
        IDToneInterval.M7,
        IDToneInterval.O,
    ]
    """Defines the intervals for the natural major scale."""

    HARMONIC = [
        IDToneInterval.R,
        IDToneInterval.M2,
        IDToneInterval.m3,
        IDToneInterval.P4,
        IDToneInterval.P5,
        IDToneInterval.m6,
        IDToneInterval.M7,
        IDToneInterval.O,
    ]
    """Defines the intervals for the harmonic minor scale."""

    MELODIC = [
        IDToneInterval.R,
        IDToneInterval.M2,
        IDToneInterval.m3,
        IDToneInterval.P4,
        IDToneInterval.P5,
        IDToneInterval.M6,
        IDToneInterval.M7,
        IDToneInterval.O,
    ]
    """Defines the intervals for the melodic minor scale."""


class IDScaleGenus:

    def __init__(self, type: int, name: str, intervals: list[int], mode_names: dict[int, str], order: list[int]) -> None:
        self.TYPE = type
        self.NAME = name
        self.INTERVALS = intervals
        self.MODE_NAMES = mode_names
        self.ORDER = order


class IDGenusType:
    """Defines available scale genus types."""
    DIATONIC = 0
    HARMONIC = 1
    MELODIC  = 2


SCALE_GENERA = {
    IDGenusType.DIATONIC:
    IDScaleGenus(
        IDGenusType.DIATONIC,
        "Diatonic",
        IDScaleIntervals.DIATONIC,
        {
            1: "Natural Major",
            2: "Dorian",
            3: "Phrygian",
            4: "Lydian",
            5: "Mixolydian",
            6: "Natural Minor",
            7: "Locrian",
        },
        [4, 1, 5, 2, 6, 3, 7],
    ),
    IDGenusType.HARMONIC:
    IDScaleGenus(
        IDGenusType.HARMONIC,
        "Harmonic",
        IDScaleIntervals.HARMONIC,
        {
            1: "Harmonic Minor",
            2: "Locrian Natural 6",
            3: "Major Augmented",
            4: "Ukrainian Dorian",
            5: "Phrygian Dominant",
            6: "Lydian Natural 2",
            7: "Super Locrian bb7",
        },
        [6, 3, 4, 1, 5, 2, 7],
    ),
    IDGenusType.MELODIC:
    IDScaleGenus(
        IDGenusType.MELODIC,
        "Melodic",
        IDScaleIntervals.MELODIC,
        {
            1: "Melodic Minor",
            2: "Dorian b6",
            3: "Lydian Augmented",
            4: "Lydian Dominant",
            5: "Mixolydian b6",
            6: "Locrian #2",
            7: "Super Locrian",
        },
        [3, 4, 1, 5, 2, 6, 7],
    ),
}


class IDScale:
    """Represents a scale which is defined by its tonic note and its intervals, e.g. C Natural Major."""

    def __init__(self, tonic: int | str, genus=IDGenusType.DIATONIC, mode=1) -> None:
        """
        Args:
            tonic: The root note name OR the root note value of the scale.
            genus: The IDGenusType defines the tone intervals to be used for the scale.
            mode: The mode of the scale (1-7).
        """

        self.set_tonic_note(tonic)
        self.set_mode(mode)
        self.set_genus(genus)


    def clone(self) -> 'IDScale':
        """Returns a deep copy of the chord."""
        return deepcopy(self)


    def contains(self, note_value: int) -> bool:
        """Tests if a given note value belongs to the scale."""
        return normalize_note_value(note_value) in normalize_note_values(self.note_values())


    def genus(self) -> int:
        return self._template.TYPE
    
    def genus_name(self) -> str:
        """Returns the name of the genus of the scale, i.e. 'Diatonic', 'Harmonic' or 'Melodic'."""
        return self._template.NAME
    

    def mode_name(self) -> str:
        """Returns the name of the scale's mode. e.g. 'Natural Major' or 'Mixolydian'."""
        return self._template.MODE_NAMES[self._mode]


    def mode(self) -> int:
        """Returns the mode of the scale."""
        return self._mode


    def name(self, style="flat") -> str:
        """Returns the name of the scale, e.g. 'C Natural Major' or 'F Lydian'."""
        return note_name(self._tonic_note_value, style, include_octave=False) + " " + self.mode_name()


    def note_name_in_scale(self, note_value, style="flat") -> str:
        """Returns the relative nota name of the scale for a given note value.
        
        The relative note name of the tonic of the scale is '1', the next note in the scale
        has the relative note name '2', etc. Chromatic notes (not in scale) are named e.g.
        'b1' or '#5'.

        Args:
            style (optional): 'flat' OR 'sharp', defines if 'b' or '#' shall be prioritized for chromatic notes.
        """
        sharp = (note_value - 1) % IDToneInterval.OCTAVE
        note = (note_value + 0) % IDToneInterval.OCTAVE
        flat = (note_value + 1) % IDToneInterval.OCTAVE

        normalized_scale_values = normalize_note_values(self.note_values())
        note_value_to_pos = {v: i + 1 for i, v in enumerate(normalized_scale_values)}

        if (note in normalized_scale_values):
            return str(note_value_to_pos[note])
    
        flat_in_scale = flat in normalized_scale_values
        flat_name = 'b' + str(note_value_to_pos[flat]) 

        sharp_in_scale = sharp in normalized_scale_values
        sharp_name = '#' + str(note_value_to_pos[sharp])

        if (flat_in_scale and sharp_in_scale):
            if (style == "flat"):
                return flat_name
            else:
                return sharp_name

        if (flat_in_scale):
            return flat_name
        
        if (sharp_in_scale):
            return sharp_name
        
        raise RuntimeError("Could not determine note name in scale!")



    def note_values(self, base_note_value=0, include_perfect_octave=False):
        """Returns the note values from one octave of the scale.
        
        Args:
            base_note_value (optional): This must be a C-note value. The returned values of the scale
              will be greater than this C-note value but also as close as possible to this C-note value.
            include_perfect_octave (optional): Seven notes will be returned When this parameter is False.
              Eight notes will be returned when this parameter is True.
        
        Raises:
            ValueError if input conditions are not met.
        """
        if base_note_value % IDToneInterval.OCTAVE != 0:
            raise ValueError("The base note must be a C note!")

        # calculate interval step size for the current mode
        scale_steps = deque([b - a for a, b in zip(self._template.INTERVALS[:-1], self._template.INTERVALS[1:])])
        scale_steps.rotate(1 - self._mode)

        # calculate note values realtive to the tonic
        relative_note_values = [0, *accumulate(scale_steps)]
        note_values = [v + base_note_value + self._tonic_note_value for v in relative_note_values]

        if (not include_perfect_octave):
            note_values = note_values[:-1]

        return note_values


    def set_mode(self, mode: int) -> None:
        """Sets the mode of the scale.
        
        Args:
            mode: The mode of the scale (1-7).
        """
        if (mode < 1 or mode > 7):
            raise ValueError("The mode must be in the interval 1-7!")

        self._mode = mode


    def set_genus(self, genus: int) -> None:
        """Sets the genus of the scale.
        
        Args:
            genus: A value from IDGenusType, i.e. DIATONIC, HARMONIC or MELODIC.
        """
        self._template = SCALE_GENERA[genus]

        if not self._template:
            raise ValueError("Invalid scale genus!")
        

    def set_tonic_note(self, tonic: int | str) -> None:
        """Sets the tonic note of the scale."""
        self._tonic_note_value = note_to_note_value(tonic)

    
    def tonic_note_value(self) -> int:
        """Returns the tonic note of the scale."""
        return self._tonic_note_value


    def triad_chords(self) -> list[IDChord]:
        result: list[IDChord] = []
        scale_note_values = self.note_values() + transpose_tone_intervals(self.note_values(), IDToneInterval.OCTAVE)

        for i, note in enumerate(self.note_values()):
            chord_note_values = [scale_note_values[i + IDScaleDegree.TONIC], 
                                 scale_note_values[i + IDScaleDegree.MEDIANT], 
                                 scale_note_values[i + IDScaleDegree.DOMINANT]]
            chord_tonic = chord_note_values[0]
            chord_intervals = transpose_tone_intervals(chord_note_values, -chord_tonic)

            for chord_type, chord_template in CHORD_TYPES.items():
                if chord_intervals == chord_template.intervals():
                    result.append(IDChord(chord_tonic, chord_type))


        return result


    def __str__(self) -> str:
        return self.name() + " [" + ", ".join(note_values_to_note_names(self.note_values())) + " ]"
    

    def __repr__(self) -> str:
        return self.__str__()
        



"""
This module defines basic tone intervals and some functions which can operate on them.
"""

__author__ = "https://codeberg.org/decoherence"

import itertools

class IDToneInterval:
    """Defines names and values for tone intervals."""
    # Interval definitions
    ROOT           =  0
    TONIC          =  0
    SEMI_TONE      =  1
    TONE           =  2 * SEMI_TONE
    MINOR_2ND      =  1 * SEMI_TONE
    MAJOR_2ND      =  2 * SEMI_TONE
    MINOR_3RD      =  3 * SEMI_TONE
    MAJOR_3RD      =  4 * SEMI_TONE
    PERFECT_4TH    =  5 * SEMI_TONE
    DIMINISHED_5TH =  6 * SEMI_TONE
    AUGMENTED_4TH  =  6 * SEMI_TONE
    TRITONE        =  6 * SEMI_TONE
    PERFECT_5TH    =  7 * SEMI_TONE
    AUGMENTED_5TH  =  8 * SEMI_TONE
    MINOR_6TH      =  8 * SEMI_TONE
    MAJOR_6TH      =  9 * SEMI_TONE
    MINOR_7TH      = 10 * SEMI_TONE
    MAJOR_7TH      = 11 * SEMI_TONE
    PERFECT_OCTAVE = 12 * SEMI_TONE
    MINOR_9TH      = 13 * SEMI_TONE
    MAJOR_9TH      = 14 * SEMI_TONE
    PERFECT_11TH   = 17 * SEMI_TONE
    MAJOR_13TH     = 21 * SEMI_TONE
    OCTAVE         = PERFECT_OCTAVE

    # Abbreviations
    P1 = TONIC
    R  = ROOT
    m2 = MINOR_2ND
    M2 = MAJOR_2ND
    m3 = MINOR_3RD
    M3 = MAJOR_3RD
    P4 = PERFECT_4TH
    dim5 = DIMINISHED_5TH
    Aug4 = AUGMENTED_4TH
    T = TRITONE    
    P5 = PERFECT_5TH    
    Aug5 = AUGMENTED_5TH
    m6 = MINOR_6TH    
    M6 = MAJOR_6TH
    m7 = MINOR_7TH
    M7 = MAJOR_7TH
    P8 = PERFECT_OCTAVE
    m9 = MINOR_9TH
    M9 = MAJOR_9TH
    P11 = PERFECT_11TH
    M13 = MAJOR_13TH
    O = OCTAVE

_TONE_INTERVALS_LONG_NAMES = {
    IDToneInterval.R:    "Root", 
    IDToneInterval.m2:   "minor 2nd",
    IDToneInterval.M2:   "Major 2nd",
    IDToneInterval.m3:   "minor 3rd",
    IDToneInterval.M3:   "Major 3rd",
    IDToneInterval.P4:   "Perfect 4th",
    IDToneInterval.dim5: "Tritone",
    IDToneInterval.P5:   "Perfect 5th",
    IDToneInterval.m6:   "minor 6th",
    IDToneInterval.M6:   "Major 6th",
    IDToneInterval.m7:   "minor 7",
    IDToneInterval.M7:   "Major 7",
    IDToneInterval.O:    "Octave",
    IDToneInterval.m9:   "minor 9th",
    IDToneInterval.M9:   "Major 9th",
    IDToneInterval.P11:  "Perfect 11th",
    IDToneInterval.M13:  "Major 13th"
    }

_TONE_INTERVALS_SHORT_NAMES = {
    IDToneInterval.R:    "R", 
    IDToneInterval.m2:   "m2",
    IDToneInterval.M2:   "M2",
    IDToneInterval.m3:   "m3",
    IDToneInterval.M3:   "M3",
    IDToneInterval.P4:   "P4",
    IDToneInterval.dim5: "dim5",
    IDToneInterval.P5:   "P5",
    IDToneInterval.m6:   "m6",
    IDToneInterval.M6:   "M6",
    IDToneInterval.m7:   "m7",
    IDToneInterval.M7:   "M7",
    IDToneInterval.O:    "O",
    IDToneInterval.m9:   "m9",
    IDToneInterval.M9:   "M9",
    IDToneInterval.P11:  "P11",
    IDToneInterval.M13:  "M13"
    }


def tone_interval_long_name(interval: int):
    return _TONE_INTERVALS_LONG_NAMES[interval]

def tone_interval_short_name(interval: int):
    return _TONE_INTERVALS_SHORT_NAMES[interval]


def normalize_tone_intervals(intervals: list[int] | set[int]) -> list[int]:
    """Normalizes a list of tone intervals to be in the range ROOT (0) to MAJOR_7TH (11)."""
    normalized_set = {v % IDToneInterval.OCTAVE for v in intervals}
    return sorted(normalized_set)


def transpose_tone_intervals(intervals: list[int], semi_tones: int) -> list[int]:
    """Transposes the given tone intervals a number of semi tone steps"""
    return [value + semi_tones for value in intervals if (value + semi_tones) >= 0]


def tone_intervals_signature(intervals: list[int]) -> int:
    """Translates a list of tone intervals to an interger number which is unique for the set of normalized intervals.

    Each bit in the signature represents a note in the normalized (see normalizeIntervals) interval:
        0: note not present
        1: note present
    
    Returns:
        A normalzed intervals signature number between 0 and 4095 (2^12 - 1).
    """
    signature = 0
    mask = 1
    
    for value in normalize_tone_intervals(intervals):
        signature = signature | (mask << value)
        
    return signature


def near_tone_intervals_signatures(signature: int, distance: int) -> list[int]:
    """Finds normalized intervals signatures which are close to a given signature
    
    Args:
        signature: The target signature (see toneIntervalsSignature).
        distance: The function will return signatures which are at this distance from the target signature.
          The distance is defined as number of notes which differs from the target signature.
    
    """
    near_signatures = []

    if distance < 0:
        raise ValueError("Distance must be positive or zero!")

    if distance == 0:
        return [signature]
    
    bits_to_toggle = [list(t) for t in itertools.combinations(range(IDToneInterval.OCTAVE), distance)]

    for bits in bits_to_toggle:
        mask = 0
        for bit in bits:                
            mask = mask | (1 << bit)

        near_signatures.append(signature ^ mask) # xor toggles the bits in the mask

    return near_signatures





"""
This module defines special characters which are used when expressing musical terms.
"""


__author__ = "https://codeberg.org/decoherence"


class IDMusicalChar:

    DIM = chr(176)
    """The 'degree' symbol used for diminished chords."""

    FLAT = chr(9837)
    """The 'b' symbol used for flat notes and chords."""

    HALF_DIM = chr(119209)
    """The 'slashed degree' symbol used for half-diminished chords."""


"""
This module defines chords and functions for handling chords.
"""

__author__ = "https://codeberg.org/decoherence"

import copy
from itertools import combinations
from functools import reduce

from idroman import integer_to_roman

from .characters import IDMusicalChar
from .notes import note_name, note_to_note_value, generated_note_values
from .tone_intervals import IDToneInterval, tone_intervals_signature, near_tone_intervals_signatures


class IDScaleDegree:
    TONIC = 0
    SUPERTONIC = 1
    MEDIANT = 2
    SUBDOMINANT = 3
    DOMINANT = 4
    SUBMEDIANT = 5
    SUBTONIC = 6


SCALE_DEGREE_NAMES = {
    IDScaleDegree.TONIC: "Tonic", 
    IDScaleDegree.SUPERTONIC: "Supertonic", 
    IDScaleDegree.MEDIANT: "Mediant", 
    IDScaleDegree.SUBDOMINANT: "Subdominant", 
    IDScaleDegree.DOMINANT: "Dominant", 
    IDScaleDegree.SUBMEDIANT: "Submediant", 
    IDScaleDegree.SUBTONIC: "Subtonic"
    }
"""Mapping between relative note position in a scale and the name of the scale degree name."""


class _IDChordMod:
    """An IDChordModifier can be applied to a IDChord to alter its notes, e.g. to change a G chord to a G7 chord."""

    def __init__(self, long_name: str, short_name: str, to_add: list[int], to_remove: list[int], cancels: list[int]) -> None:
        """
        Args:
            long_name: The long name of the modifier, e.g. 'dominant 7'.
            short_name: The short name of the modifier, e.g. '7'.
            to_add: A list of relative intervals to be added to the chord, e.g. '[IDToneInterval.m7]'.
            to_remove: A list of relative intervals to be removed from the chord.
            cancels: A list of relative intervals which represents a modifier which shall be cancelled by this modifier.            
        """

        self._long_name = long_name
        self._short_name = short_name
        self._to_add = to_add
        self._to_remove = to_remove
        self._cancels = cancels

    
    def apply(self, tonic: int, intervals: list[int]) -> list[int]:
        """Applies this modifier to given input intervals and returns the result.
        
        All 'to_add' intervals are added and all 'to_remove' are removed from given 'intervals'.

        Args:
            tonic_note_value: The root note value to be used with 'source'.
            intervals: The input interval values
        """
        removing = [tonic + i for i in self._to_remove]
        adding   = [tonic + i for i in self._to_add]

        result = [v for v in intervals if v not in removing]
        result.extend(adding)
        result.sort()

        return result
    

    def append_short_name(self, input: str) -> str:
        """Returns the input string with the short name of the modifier appended."""
        return input + self._short_name
    

    def append_long_name(self, input: str) -> str:
        """Returns the input string with the long name of the modifier appended."""

        # Fix to avoid getting 'major major 7'
        if (" major" in input) and ("major 7" not in input):
            input = input.replace(" major", "")

        return input + " " + self._long_name
    

    def cancels_modifiers(self) -> list[int]:
        """Returns a list with the modifiers which shall be cancelled by this modifier."""
        return self._cancels
    

    def long_name(self) -> str:
        """Returns the long name of the modifier, e.g. 'dominant 7'."""
        return self._long_name
    

    def short_name(self) -> str:
        """Returns the short name of the modifier, e.g. '7'."""
        return self._short_name


    def __str__(self) -> str:
        """Enables print of IDChordModifier."""
        return f"IDChordMod({self._short_name})"
    

    def __repr__(self) -> str:        
        """Enables print of IDChordModifier."""
        return self.__str__()

    
class IDChordFlags:
    """Flags which identifies different chord modifiers; can be added toghether with the binary | operator."""
    NO_FLAG       = 0b0000000000000000
    DOMINANT_7    = 0b0000000000000001
    MAJOR_7       = 0b0000000000000010
    DOMINANT_9    = 0b0000000000000100
    DOMINANT_11   = 0b0000000000001000
    DOMINANT_13   = 0b0000000000010000
    ADD_2         = 0b0000000000100000
    ADD_6         = 0b0000000001000000
    ADD_9         = 0b0000000010000000
    FLAT_5TH      = 0b0000000100000000
    SUSPENDED_2ND = 0b0000001000000000
    SUSPENDED_4TH = 0b0000010000000000


CHORD_MODIFIERS = {
    IDChordFlags.DOMINANT_7:    _IDChordMod("dominant 7",    "7",    
                                            to_add=[IDToneInterval.m7], 
                                            to_remove=[],
                                            cancels=[]),

    IDChordFlags.MAJOR_7:       _IDChordMod("major 7",       "maj7", 
                                            to_add=[IDToneInterval.M7], 
                                            to_remove=[],
                                            cancels=[]),

    IDChordFlags.DOMINANT_9:    _IDChordMod("dominant 9",    "9",    
                                            to_add=[IDToneInterval.m7, IDToneInterval.M9], 
                                            to_remove=[], 
                                            cancels=[IDChordFlags.DOMINANT_7]),

    IDChordFlags.DOMINANT_11:   _IDChordMod("dominant 11",   "11",   
                                            to_add=[IDToneInterval.m7, IDToneInterval.M9, IDToneInterval.P11], 
                                            to_remove=[], 
                                            cancels=[IDChordFlags.DOMINANT_7, IDChordFlags.DOMINANT_9]),

    IDChordFlags.DOMINANT_13:   _IDChordMod("dominant 13",   "13",   
                                            to_add=[IDToneInterval.m7, IDToneInterval.M9, IDToneInterval.P11, IDToneInterval.M13], 
                                            to_remove=[],
                                            cancels=[IDChordFlags.DOMINANT_7, IDChordFlags.DOMINANT_9, IDChordFlags.DOMINANT_11]),

    IDChordFlags.SUSPENDED_2ND: _IDChordMod("suspended 2nd", "sus2", 
                                            to_add=[IDToneInterval.M2], 
                                            to_remove=[IDToneInterval.m3, IDToneInterval.M3],
                                            cancels=[]),

    IDChordFlags.SUSPENDED_4TH: _IDChordMod("suspended 4th", "sus4", 
                                            to_add=[IDToneInterval.P4], 
                                            to_remove=[IDToneInterval.m3, IDToneInterval.M3],
                                            cancels=[]),

    IDChordFlags.ADD_2:         _IDChordMod("add 2",         "+2", 
                                            to_add=[IDToneInterval.M2], 
                                            to_remove=[],
                                            cancels=[]),

    IDChordFlags.ADD_6:         _IDChordMod("add 6",         "+6",    
                                            to_add=[IDToneInterval.M6], 
                                            to_remove=[],
                                            cancels=[]),

    IDChordFlags.ADD_9:         _IDChordMod("add 9",         "+9", 
                                            to_add=[IDToneInterval.M9], 
                                            to_remove=[],
                                            cancels=[]),

    IDChordFlags.FLAT_5TH:      _IDChordMod("flat 5th",     "b5",   
                                            to_add=[IDToneInterval.dim5], 
                                            to_remove=[IDToneInterval.P5],
                                            cancels=[]),

}
"""Defines dictionary dict[IDChordModFlag, IDChordModifier] with available chord modifiers."""


def list_of_chord_flags_to_chord_flags(flags: list[int]) -> int:
    """Converts from a list of IDChordFlags to a combination of IDChordFlags as a single number."""
    return reduce(lambda acc, f: acc | f, flags, IDChordFlags.NO_FLAG)


def chord_flags_to_list_of_chord_flags(flags: int) -> list[int]:
    """Converts from a combination of IDChordFlags to a list of IDChordFlags."""
    return [f for f in CHORD_MODIFIERS.keys() if (f & flags) ]


class IDChordTemplate:
    """A template for different triad chord types, e.g. major, minor etc."""

    def __init__(self, long_name: str, short_name: str, intervals: list[int]) -> None:
        """
        Args:
            long_name: The long name of the chord type, e.g. 'minor'.
            short_name: The short name of the chord type, e.g. 'm'.
            intervals: A list of relative interval values which defines the chord type.
        """
        self._long_name = long_name
        self._short_name = short_name
        self._intervals = intervals


    def intervals(self) -> list[int]:
        return self._intervals


    def note_values(self, tonic: int) -> list[int]:
        """Returns the note values of the chord.
        
        Args:
            root: The root note value of the chord.
        """
        return [n + tonic for n in self._intervals]
    

    def short_name(self, tonic:int, style="flat") -> str:
        """Returns the short name of the chord, e.g. 'Cm'."""

        return note_name(tonic, style, include_octave=False) + self._short_name
    

    def short_type_name(self) -> str:
        """Retirns the short name of the chord type, e.g. 'm' for minor"""
        return self._short_name


    def long_type_name(self) -> str:
        """Retirns the long name of the chord type, e.g. 'minor'"""
        return self._long_name


    def long_name(self, tonic: int, style="flat") -> str:
        """Returns the long name of the chord, e.g. 'C minor'."""

        if len(self._long_name) > 0:
            space = " "
        else:
            space = ""

        return note_name(tonic, style, include_octave=False) + space + self._long_name


class IDChordType:
    """Definition of values representing the different types of triad chords."""
    MAJOR      = 1
    MINOR      = 2
    DIMINISHED = 3
    AUGMENTED  = 4 


CHORD_TYPES = {
    IDChordType.MAJOR:      IDChordTemplate("major",      "",                 [IDToneInterval.R, IDToneInterval.M3, IDToneInterval.P5]),
    IDChordType.MINOR:      IDChordTemplate("minor",      "m",                [IDToneInterval.R, IDToneInterval.m3, IDToneInterval.P5]),
    IDChordType.DIMINISHED: IDChordTemplate("diminished", IDMusicalChar.DIM,  [IDToneInterval.R, IDToneInterval.m3, IDToneInterval.dim5]),
    IDChordType.AUGMENTED:  IDChordTemplate("augmented",  "+",                [IDToneInterval.R, IDToneInterval.M3, IDToneInterval.Aug5])}
"""Dictionary dict[IDChordType, IDChordTemplate] which defines templates for avaliable triad chord types."""


def chord_scale_degree_to_roman(chord_type: int, scale_degree: int) -> str:
    """Returns the roman number symbol for given IDChordType and IDScaleDegree."""
    roman_case = {IDChordType.MAJOR: "upper", 
                  IDChordType.MINOR: "lower", 
                  IDChordType.DIMINISHED: "lower", 
                  IDChordType.AUGMENTED: "lower"}
    
    post_fix = {IDChordType.MAJOR: "", 
                IDChordType.MINOR: "", 
                IDChordType.DIMINISHED: IDMusicalChar.DIM, 
                IDChordType.AUGMENTED: "+"}

    return integer_to_roman(scale_degree + 1, roman_case[chord_type]) + post_fix[chord_type]
    


class IDChord:
    """Represents a chord which can be modified by applying IDChordMods."""

    def __init__(self, tonic: int | str, type: int, flags: int | list[int] = IDChordFlags.NO_FLAG) -> None:
        """
        Args:
            tonic: Root note value OR root note name of the chord.
            type: The IDChordType which defines the type of the chord.
            flags (optional): A list of IDChordFlags which each represents a chord modifier to be applyed, OR
              an integer with accumulated IDChordFlags, e.g. 'IDChordFlags.DOMINANT_7 | IDChordFlags.ADD_9'.
        """
        self._tonic = note_to_note_value(tonic)
        self._type = type
        self._flags: list[int] = []    
        self._inversion = 0 
    
        self.set_flags(flags)


    def clone(self) -> 'IDChord':
        """Returns a deep copy of the chord."""

        return copy.copy(self)
    

    def cycle_inversion(self):
        """Increases the inversion by one, module number of notes of the chord."""
        self.set_inversion(self._inversion + 1)


    def flags(self) -> int:
        """Returns the current modifications of the chord as a combination of IDChordFlags."""
        return list_of_chord_flags_to_chord_flags(self._flags)


    def harmony(self) -> set[int]:
        """Returns the harmony of the chord as a set of note values.
        
        The tonic note will be in octave 0.
        Selected inversion will be applied (see set_inversion).
        """
        return set(self.note_values())


    def inversion(self) -> int:
        """Returns the current inversion (0-7) of the chord."""
        return self._inversion


    def long_name(self, style="flat") -> str:
        """Returns the full long name of the chord including modifiers."""
        name = self._template().long_name(self._tonic, style)

        for flag in self._flags:
            name = CHORD_MODIFIERS[flag].append_long_name(name)

        return name


    def note_values(self) -> list[int]:
        """Returns the note values of the chord.
        
        The tonic note will be in octave 0.
        Selected inversion will be applied (see set_inversion).
        """
        values = self._template().note_values(self._tonic)

        # apply modifiers
        for flag in self._flags:            
            values = CHORD_MODIFIERS[flag].apply(self._tonic, values)

        # apply inversion        
        for _ in range(self._inversion):
            values = [values[-1] - IDToneInterval.OCTAVE] + values[:-1]            
            if values[0] < 0:
                values = [i + IDToneInterval.OCTAVE for i in values]

        return values


    def number_of_flags(self) -> int:
        return len(self._flags)


    def number_of_notes(self) -> int:
        """Returns the number of notes of the chord."""
        return len(self.note_values())


    def set_flags(self, flags: int | list[int]):
        """Sets the modifications which shall be applied to the chord.
        
        Args:
            flags: A list of IDChordFlags which each represents a chord modifier to be applyed, OR
            an integer with accumulated IDChordFlags, e.g. 'IDChordFlags.DOMINANT_7 | IDChordFlags.ADD_9'.
        """
        temp_flags = []

        if (isinstance(flags, list)):
            temp_flags = [f for f in flags if f != IDChordFlags.NO_FLAG]

        if (isinstance(flags, int)):
            if (flags != IDChordFlags.NO_FLAG):        
                temp_flags = chord_flags_to_list_of_chord_flags(flags)

        flags_to_be_canceled = {f for g in temp_flags for f in CHORD_MODIFIERS[g].cancels_modifiers()}
        self._flags = [f for f in temp_flags if f not in flags_to_be_canceled]
        

    def set_inversion(self, steps: int):
        """Sets the inversion of the chord.

        When a chord is inverted, the root note will no longer be the lowest note of the chord.
        For each inversion step, another of the notes of the chord will be the lowest note
        of the chord.
        
        Args:
            steps: The number of inversion steps. Inversion steps are modulo N, where N is
              the number of notes of the chord, i.e. inversion N + 1 = inversion 0.            
        """
        self._inversion = steps % self.number_of_notes()


    def set_tonic(self, tonic: int | str):
        """Sets the root note of the chord.
        
        Args:
            root: Note value OR note name.
        """
        self._tonic = note_to_note_value(tonic)


    def set_type(self, type: int):
        """Sets the type of the chord.

        Args:
            type: The IDChordType which defines the type of the chord.
        """
        self._type = type


    def short_mod_name(self, style="flat") -> str:
        """Returns the combined short name of the modifiers and inversion, without the tonic note name and the chord type, e.g. '7add9/G'."""
        name = ""

        for flag in self._flags:
            name = CHORD_MODIFIERS[flag].append_short_name(name)

        base_note_value = self.note_values()[0]
        if base_note_value != self._tonic:
            name = name + "/" + note_name(base_note_value, style, include_octave=False)

        return name


    def short_name(self, style="flat") -> str:
        """Returns the full short name of the chord including modifiers and inversion."""
        return self.short_type_name(style) + self.short_mod_name()


    def short_type_name(self, style="flat") -> str:
        """Returns the short name of the chord without modifiers, i.e. 'C#m' for 'minor'."""
        return self._template().short_name(self._tonic, style)


    def signature(self) -> int:
        return tone_intervals_signature(self.note_values())


    def tonic_note_value(self) -> int:
        """Returns the root note value of the chord."""
        return self._tonic
    

    def type(self) -> int:
        """Returns the IDChordType of the chord."""
        return self._type


    def _template(self) -> IDChordTemplate:
        return CHORD_TYPES[self._type]
    

    def __eq__(self, other) -> bool:
        """Compare operator for IDChord.
        
        Args:
            other: IDChord.
        """

        if isinstance(other, IDChord):
            return self.note_values() == other.note_values()
        
        if isinstance(other, list):
            return self.note_values() == other
                
        raise ValueError("Invalid type!")
    

    def __hash__(self):
        return hash(tuple(self.note_values())) 
    

    def __ne__(self, other) -> bool:
        """Compare operator for IDChord."""
        return not self.__eq__(other)


    def __str__(self):
        """Enables print of IDChord."""
        return f"IDChord({self.long_name()} | {self.short_name()})"
    

    def __repr__(self):
        """Enables print of IDChord."""
        return self.__str__()
    

class IDChordDatabase:
    """An instance of this class is a database with chords of all types, all normalized root notes and 
    combinations of chord modifications.    
    """

    def __init__(self, number_mod_combinations = 3) -> None:
        """
        Args:
            number_mod_combinations (optional): An integer which defines the number of combinations of
              chord modifications will be applied when the database is created.
        """
        
        self._chord_database: dict[int, set[IDChord]] = dict()
        """The chord database is a directory with the chord signature as key value. Since different
        chords can have the same signature, each database entry may contain several chords.
        """

        print("Creating chord database ...")

        all_chord_flags = [IDChordFlags.NO_FLAG, *CHORD_MODIFIERS.keys()]
        chord_flag_combinations = [list(c) for c in combinations(all_chord_flags, number_mod_combinations)]
        
        for tonic in generated_note_values("C", IDToneInterval.OCTAVE):
            for type in CHORD_TYPES.keys():
                self._add_chord(IDChord(tonic, type, IDChordFlags.NO_FLAG))

                for flags in chord_flag_combinations:
                    self._add_chord(IDChord(tonic, type, flags))


        print(f"  Chords in database: {self.size()}")
        print(f"  Unique chord signatures in database: {len(self._chord_database)}")


    def _add_chord(self, chord: IDChord) -> None:
        signature = chord.signature()

        if (not signature in self._chord_database.keys()):
            self._chord_database[signature] = set()
        signature_entry = self._chord_database[signature]

        # remove chords with identical note values with more flags or more complex flags
        for existing_chord in signature_entry:
            if (chord == existing_chord):
                if (chord.number_of_flags() < existing_chord.number_of_flags()):
                    signature_entry.remove(existing_chord)
                    break
                elif (chord.flags() < existing_chord.flags()):                    
                    signature_entry.remove(existing_chord)
                    break

        signature_entry.add(chord)


    def size(self) -> int:
        """Returns the number of chords in the database."""
        result = 0

        for chords in self._chord_database.values():
            result += len(chords)

        return result
    

    def match(self, intervals: list[int], distance: int = 0) -> list[IDChord]:
        """Returns chords found in the database which matches the input intervals.
        
        Args:
            intervals: The input intervals are normalized and compared with the
              normalized intervals of the chords in the database.
            distance: The number of notes which shall differ to make a match, e.g.
              distance = 0 returns exact matches, distance = 1 returns chords which
              differs with one note.
        """
        chords: list[IDChord] = []
        input_signature = tone_intervals_signature(intervals)        
        signatures_to_seach_for = near_tone_intervals_signatures(input_signature, distance)        


        for signature in signatures_to_seach_for:
            if signature in self._chord_database:
                chords.extend(self._chord_database[signature])

        return chords




"""
This module defines note values, note names and functions for handling notes.
"""

__author__ = "https://codeberg.org/decoherence"


import re
from string import digits
from typing import List, Union
from .tone_intervals import IDToneInterval


NOTE_NAMES_TEMPLATE_SHARP = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
NOTE_NAMES_TEMPLATE_FLAT  = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"]


def note_name_style(list_of_note_names: list[str]) -> str:
    """Returns the style (sharp/flat) used for given note names.
    
    If the style cannot be determined (no 'b' or '#' in anly of the note names),
    the default style 'sharp' is assumed.

    Returns: 'sharp' or 'flat'
    """
    if any('b' in name for name in list_of_note_names):
        return "flat"
    else:
        return "sharp"
    

def note_name(note_value: int, style="flat", include_octave=True) -> str:
    """Returns the note name for a given note value.
    
    Args:
        note_value: The lowest note name 'C0' has note value 0; other notes have note values which represents
          their distance from 'C0' measured in number semi tones.
        style (optional): 'sharp' or 'flat'
        show_octave (optional): Indicates if the octave number shall be included in the note name.
    
    """
    match style:
        case "sharp":
            note_names_template = NOTE_NAMES_TEMPLATE_SHARP
        case "flat":
            note_names_template = NOTE_NAMES_TEMPLATE_FLAT
        case _:
            raise ValueError("Style must be 'sharp' or 'flat'")

    base_name_index = note_value % IDToneInterval.OCTAVE

    post_fix = ""
    if include_octave:
        post_fix = str(note_value // IDToneInterval.OCTAVE)
    
    return note_names_template[base_name_index] + post_fix


def note_value(note_name: str) -> int:
    """Returns the note value for a given note name.
    
    Raises:
        ValueError if the input in not a valic note name.
    
    """
    style = note_name_style([note_name])

    match style:
        case "sharp":
            name_template = NOTE_NAMES_TEMPLATE_SHARP
        case "flat":
            name_template = NOTE_NAMES_TEMPLATE_FLAT
        case _:
            raise ValueError("Style must be 'sharp' or 'flat'")

    pattern = re.compile(r'(\D+)(\d*)')
    match = pattern.search(note_name)

    base_note_name = ""
    octave_str = ""

    if (match):
        base_note_name = match.group(1)
        octave_str = match.group(2)
        
    try:
        octave = int(octave_str)
    except:
        octave = 0

    return name_template.index(base_note_name) + octave * IDToneInterval.OCTAVE


def note_to_note_value(note: str | int) -> int:
    """Translates a note value OR a note name to its note value.
    
    Raises:
        ValueError if the input cannot be decoded to a note value.
    """
    if isinstance(note, str):
        return note_value(note)
    elif isinstance(note, int):
        return note
    else:
        raise ValueError("Parameter must be of type str or int")


def normalize_note_value(note_value: int) -> int:
    return note_value % IDToneInterval.OCTAVE


def normalize_note_values(note_values: list[int]) -> list[int]:
    return [v % IDToneInterval.OCTAVE for v in note_values]


def generated_note_names(start_note: int | str, number_of_notes, style="flat", include_octave=True):
    """Generator for sequential note names in a given interval of notes.

    Args:
        start_note: Value or name of the first note.
        number_of_notes: The number of semi notes which will be generated.
        style (optional): 'sharp' or 'flat'
        include_octave (optional): Indicates if the octave number shall be included in the note name.

    Raises:
        ValueError if incorrect input.

    """
    start_note_value = note_to_note_value(start_note)

    for i in range(number_of_notes):
        yield note_name(i + start_note_value, style, include_octave)


def generated_note_values(start_note: int | str, number_of_notes):
    """Generator for sequential note values in a given interval of notes.

    Args:
        start_note: Value or name of the first note.
        number_of_notes: The number of semi notes which will be generated.        

    Raises:
        ValueError if incorrect input.

    """
    start_note_value = note_to_note_value(start_note)

    for i in range(number_of_notes):
        yield i + start_note_value


def list_of_note_names(startNote: int | str, numberNotes, style="flat", include_octave=True):
    """Returns a list of sequential note names for a given interval of notes.

    Args:
        start_note: Value or name of the first note.
        number_notes: The number of semi notes which will be generated.
        style (optional): 'sharp' or 'flat'
        include_octave (optional): Indicates if the octave number shall be included in the note name.

    Raises:
        ValueError if incorrect input.

    """
    startNoteValue = note_to_note_value(startNote)
    return [name for name in generated_note_names(startNoteValue, numberNotes, style, include_octave)]


class _NoteNameDict:
    """Implements the interface of a dictory to mimic a dict[note value, note name] for all notes."""

    def __init__(self, style="flat"):
        self.style = style

    def __getitem__(self, note_value):
        return note_name(note_value, self.style)

    def index(self, note_name):
        return note_value(note_name)


NOTE_NAMES_FLAT = _NoteNameDict("flat")
"""Represents a dictionary dict [note value, note name] for all note names with the style 'flat'."""

NOTE_NAMES_SHARP = _NoteNameDict("sharp")
"""Represents a dictionary dict [note value, note name] for all note names with the style 'sharp'."""


def sorted_note_names(list_of_note_names: list[str]) -> list[str]:
    """Returns a sorted a list of note names according to its note values."""
    return sorted(list_of_note_names, key=lambda n: note_value(n))


def note_values_to_note_names(note_values: list[int], style="flat", include_octave=False) -> list[str]:
    """Transforms a list of note values to the corresponding list of note names.
    
    Args:
        note_values: List of note values to be translated.
        style (optional): 'sharp' or 'flat'
        include_octave (optional): Indicates if the octave number shall be included in the note name.
    """
    return [note_name(note_value, style, include_octave) for note_value in note_values]


def note_names_to_note_values(note_names: list[str]) -> list[int]:
    """Translates a list of note names to a list of note values."""
    return [note_value(note_name) for note_name in note_names]


def note_name_without_octave(note_name: str):
    """Strips any existing octave number from provided note name."""
    remove_digits = str.maketrans('', '', digits)
    return note_name.translate(remove_digits)


def is_diatonic_note_name(note_name: str) -> bool:
    """Returns True if the provided note name does not include any 'b' or '#'."""
    return ('b' not in note_name) and ('#' not in note_name)


def is_diatonic_note_value(note_value: int) -> bool:
    """Returns True if the provided note value does not represent a note name which includes 'b' or '#'."""
    return is_diatonic_note_name(note_name(note_value))


def _rebase_note_values(note_values: list[int], base_value: int, current_base_value:int) -> list[int]:

    if len(note_values) > 0:
        zero_based_note_values = [value - current_base_value for value in note_values]
        return [value + base_value for value in zero_based_note_values]
    else:
        return []
    

def _base_value(note_values: Union[List[int], List[List[int]]]) -> int:
    """Finds the C-note base value for given note values.
    
    The C-note base is the C-note which is closest to to the given note values but 
    also has a lower value than all given note values.
    """
    if isinstance(note_values[0], int):
        min_value = min(note_values)
    
    if isinstance(note_values[0], list):
        min_value = min(min(value_list) for value_list in note_values if value_list) #type: ignore (caller to ensure correct types)
    
    return (min_value // IDToneInterval.OCTAVE) * IDToneInterval.OCTAVE #type: ignore (caller to ensure correct types)
    

def rebase_note_values(note_values: list[int] | list[list[int]], base_value: int) -> list[int] | list[list[int]]:
    """Transposes the provided note values N octaves to be just above the given C-note.
    
    Args:
        note_values: A list of note values OR a list of lists of note values.
        base_value: A C-note value which represents the octave to which the notes shall be transposed.
          The new note values will be above the given C-note and also as close as possible to the 
          given C-note.

    Returns:
        A list on the same format as the input with transposed note values.
    
    """
    
    if len(note_values) == 0:
        return []

    if base_value % IDToneInterval.OCTAVE != 0:
        raise ValueError("Argument base_value must be a C note!")

    if isinstance(note_values[0], int):
        current_base = _base_value(note_values)
        return _rebase_note_values(note_values, base_value, current_base) #type: ignore (trust the caller to provide correct types)
    
    if isinstance(note_values[0], list):
        current_base = _base_value(note_values)
        note_value_list = list()

        for n in note_values:
            note_value_list.append(_rebase_note_values(n, base_value, current_base)) #type: ignore (trust the caller to provide correct types)

        return note_value_list
    
    return []



from .characters import (
    IDMusicalChar
)

from .chords import (
    IDChord,
    IDChordFlags,
    IDChordType,
    chord_flags_to_list_of_chord_flags,
    list_of_chord_flags_to_chord_flags,
    chord_scale_degree_to_roman,
    SCALE_DEGREE_NAMES,
    CHORD_MODIFIERS,
    CHORD_TYPES
)

from .notes import (
    NOTE_NAMES_TEMPLATE_SHARP,
    NOTE_NAMES_TEMPLATE_FLAT,
    NOTE_NAMES_FLAT,
    NOTE_NAMES_SHARP,
    note_name_style,
    note_name,
    note_value,
    note_to_note_value,
    generated_note_names,
    generated_note_values,
    list_of_note_names,
    sorted_note_names,
    note_values_to_note_names,
    note_names_to_note_values,
    note_name_without_octave,
    is_diatonic_note_name,
    is_diatonic_note_value,
    rebase_note_values,
)

from .scales import (
    SCALE_GENERA,
    IDGenusType,
    IDScaleIntervals,
    IDScale
)

from .tone_intervals import (
    IDToneInterval,
    tone_interval_long_name,
    tone_interval_short_name,
    normalize_tone_intervals,
    transpose_tone_intervals,
    tone_intervals_signature,
    near_tone_intervals_signatures,
)
